#include <QtGui>
#include "mainwindowimpl.h"
//a key press event
#define KeyPres 6
#define PRGFile 0
#define TextFile 1
#define DataFile 2
#define D64File 3
//The bam of the disk - sector allocation map
#define Track18Sector0 91392
//first sector of the directory
#define Track18Sector1 91648
//position of the disk label
#define DiskLabel 91536
//position of the bit allocation map of the directory sectors
#define BeginDirBam 91464
//end of the bit allocation map of the directory sectors
#define EndDirBam 96512
//
MainWindowImpl::MainWindowImpl( QWidget * parent, Qt::WFlags f) 
	: QMainWindow(parent, f)
{
	//various strings
	PRGHeader = "Open Commodore Program Files";
	BasicHeader = "Open Text Files";
	DataHeader = "Open Files in Hex Viewer";
	//strings for the load save file dialogs
	PRGType = "PRG Files (*.PRG *.prg);;All Files (*)";
	BasicType = "All Files (*);;ASM Files (*.ASM .*asm);;Text Files (*.TXT *.txt)";
	DataType = "All Files (*)";
	D64Header = "Open a d64 Disk File";
	D64Type = "D64 Files (*.d64 *.D64)";
	//set default values for these global and local values
	fndtxt = "";
	prst = 0;
	arrange = 0;
	int n = 0;
	int m = 0;
	int i;
	//to show that all dialog windows are inactive - used later to insure that
	//all dialogs are closed before the program ends.
	FndDlg = NULL;
	AboutDialog = NULL;
	Aboutqt = NULL;
	CfgDialog = NULL;
	InsDialog = NULL;
	fnlog = NULL;
	GotoDlg = NULL;
	d64Dialog = NULL;
	RenDlg = NULL;
	AsmResDialog = NULL;
	RenumberInProgress = false;
	//setup the global basic strings
	MdiChild mmd;
	mmd.Init();
	//get the home dir and make the string to search for the saved settings
	QDir Home;
	QString mhome =  Home.homePath();
	homedir = mhome + "/.CommodoreEditor";
	inifile = homedir + "/settings.ini";
	//if the settings file exists
	if(Home.exists(inifile))
	{
		QFile infile(inifile);
		if(infile.open(QIODevice::ReadOnly))
		{
			//get the file size and read it into a buffer
			QString inidata;
			i = infile.size();
			char* fibuf = new char[i + 2];
			infile.read(fibuf,i);
			fibuf[i] = 0;
			infile.close();
			inidata = fibuf;
			//search the buffer for the saved settings and initilize the variables
			int k = inidata.indexOf("PRGDir ") + 7;
			int j = inidata.indexOf("TxtDir ") - 1;
			//intilize the default directories for the different file types
			Prgdir = inidata.mid(k,j - k);
			k = inidata.indexOf("DataDir ") - 1;
			j += 8;
			Txtdir = inidata.mid(j,k - j);
			k += 9;
			j = inidata.indexOf("D64Dir ") - 1;
			Datadir = inidata.mid(k,j - k);
			j += 8;
			k = inidata.indexOf("AsmCmdLine ") - 1;
			D64Dir = inidata.mid(j,k - j);
			k += 12;
			j = inidata.indexOf("PRGFont ") - 1;
			//get the command line for the external assembler (normally a script)
			AsmCmdLine = inidata.mid(k,j - k);	
			j += 9;
			k = inidata.indexOf("PRGFontSize ") - 1;
			//get the default fonts for the different file types
			Prgfnt = inidata.mid(j,k - j);
			k += 13;
			j = inidata.indexOf("TxtFont ") - 1;
			Prgfntsz = inidata.mid(k,j - k);
			j += 9;
			k = inidata.indexOf("TxtFontSize ") - 1;
			Txtfnt = inidata.mid(j,k - j);
			k += 13;
			j = inidata.indexOf("DataFont ") - 1;
			Txtfntsz = inidata.mid(k,j - k);
			j += 10;
			k = inidata.indexOf("DataFontSize ") - 1;
			Datafnt = inidata.mid(j,k - j);
			k += 14;
			j = inidata.indexOf("Tabs ") - 1;
			Datafntsz = inidata.mid(k,j - k);
			//the number of spaces for a tab
			Tabspace = inidata.mid(j+6,1);
			//highlight a text line or not flag
			TxtLineHighlight = inidata.mid(j+17,1);
			//now get the recently open file list -- max 8 files
			for(j+=19;j < i;j = k + 3)
			{
				k = inidata.indexOf(QChar(10),j);
				RecentFiles[m] = inidata.mid(j,k - j);
				RecentFileType[m++] = inidata.mid(k+1,1).toInt();
			}
			for(n = m;m < 8;m++)
				RecentFiles[m] = "";
			delete fibuf;
		}
		else
		{
			//if the settings file does not exist set standard values
			Prgdir = mhome;
			Txtdir = mhome;
			Datadir = mhome;
			D64Dir = mhome;
			AsmCmdLine = "";
			TxtLineHighlight = "1";
			Prgfnt = "PET News set 1";
			Txtfnt = "Sans Serfi";
			Datafnt = "Monospace";
			Prgfntsz = "10";
			Txtfntsz = "10";
			Datafntsz = "10";
			Tabspace = "5";
		}
	}
	else
	{
		//if the settings directory does not exist then make the directory and
		//make a default settings file with just a default directory setting.
		Home.mkpath(homedir);
 		QByteArray ba = mhome.toLatin1();
 		char* buf = ba.data(); 
		QFile nfile(inifile);
		if(nfile.open(QIODevice::ReadWrite))
		{
			nfile.write(buf,(qint64)mhome.length());
			nfile.close();
		}
		//set default values
		Prgdir = mhome;
		Txtdir = mhome;
		Datadir = mhome;
		D64Dir = mhome;
		AsmCmdLine = "";
		TxtLineHighlight = "1";
		Prgfnt = "PET News set 1";
		Txtfnt = "Sans Serfi";
		Datafnt = "Monospace";
		Prgfntsz = "10";
		Txtfntsz = "10";
		Datafntsz = "10";
		Tabspace = "5";
	}
	//setup the three fonts needed
	PRGfont = QFont(Prgfnt);
	PRGfont.setPointSize(Prgfntsz.toInt());
	Txtfont = QFont(Txtfnt);
	Txtfont.setPointSize(Txtfntsz.toInt());
	Datafont = QFont(Datafnt);
	Datafont.setPointSize(Datafntsz.toInt());
	//and the tab settings
	txttab = Tabspace.toInt();
	//setup the main dialog
	setupUi(this);
	//setup the recent file list
	for(i = 0;i < n;i++)
	{
		RecentFilesAction[i] = new QAction(this);
		m = RecentFiles[i].lastIndexOf("/") + 1;
		RecentFilesAction[i]->setText(RecentFiles[i].right(RecentFiles[i].length()-m));
		RecentFilesAction[i]->setData(RecentFiles[i]);
		RecentFilesAction[i]->setVisible(true);
 		connect(RecentFilesAction[i],SIGNAL(triggered()),
 			this,SLOT(actionRecentFiles()));
 		menuRecent_Files->addAction(RecentFilesAction[i]);
	}
	//setup the column display
	QLabel* qql = new QLabel("Column");
	statusBar()->addPermanentWidget(qql);
	ColumnCursor = new QLabel("0");
	ColumnCursor->setMinimumSize(30,15);
	ColumnCursor->setFrameStyle(QFrame::StyledPanel);
	statusBar()->addPermanentWidget(ColumnCursor);
	//setup the caps - scroll - num display
	Capslock = new QLabel("CAPS");	
	Capslock->setMinimumSize(40,15);
	Capslock->setFrameStyle(QFrame::StyledPanel | QFrame::Raised);
	statusBar()->addPermanentWidget(Capslock);
	Numlock = new QLabel("NUM");	
	Numlock->setMinimumSize(40,15);
	Numlock->setFrameStyle(QFrame::StyledPanel | QFrame::Raised);
	statusBar()->addPermanentWidget(Numlock);
	Scrolllock = new QLabel("SCROL");	
	Scrolllock->setMinimumSize(40,15);
	Scrolllock->setFrameStyle(QFrame::StyledPanel | QFrame::Raised);
	statusBar()->addPermanentWidget(Scrolllock);
	//setup the controls caps - scroll - num for linux
#ifndef WIN_32
 	Display* pDisplay = XOpenDisplay(NULL);
 	Capslok = keyState(XK_Caps_Lock,pDisplay);
	if(Capslok)
		Capslock->setText("<font color='red'>CAPS</font>");
	else
		Capslock->setText("<font color='black'>CAPS</font>");
	Numlok = keyState(XK_Num_Lock,pDisplay);
	if(Numlok)
		Numlock->setText("<font color='red'>NUM</font>");
	else
		Numlock->setText("<font color='black'>NUM</font>");
	Scrolok = keyState(XK_Scroll_Lock,pDisplay);
	if(Scrolok)
		Scrolllock->setText("<font color='red'>SCROL</font>");
	else
		Scrolllock->setText("<font color='black'>SCROL</font>");
 	XCloseDisplay(pDisplay);
#endif
	statusBar()->showMessage(tr("Ready"));
}

MainWindowImpl::~MainWindowImpl()
{
	//setup the data to write the settings file
	QString inidata;
	inidata = "PRGDir ";
	inidata += Prgdir;
	inidata += QChar(10);
	inidata += "TxtDir ";
	inidata += Txtdir;
	inidata += QChar(10);
	inidata += "DataDir ";
	inidata += Datadir;
	inidata += QChar(10);
	inidata += "D64Dir ";
	inidata += D64Dir;
	inidata += QChar(10);
	inidata += "AsmCmdLine ";
	inidata += AsmCmdLine;
	inidata += QChar(10);
	inidata += "PRGFont ";
	inidata += Prgfnt;
	inidata += QChar(10);
	inidata += "PRGFontSize ";
	inidata += Prgfntsz;
	inidata += QChar(10);
	inidata += "TxtFont ";
	inidata += Txtfnt;
	inidata += QChar(10);
	inidata += "TxtFontSize ";
	inidata += Txtfntsz;
	inidata += QChar(10);
	inidata += "DataFont ";
	inidata += Datafnt;
	inidata += QChar(10);
	inidata += "DataFontSize ";
	inidata += Datafntsz;
	inidata += QChar(10);
	inidata += "Tabs ";
	inidata += Tabspace;
	inidata += QChar(10);
	inidata += "TxtHLite ";
	inidata += TxtLineHighlight;
	inidata += QChar(10);
	//get the recent files
	QString astr;
	for(int i = 0;i < 8;i++)
	{
		if(RecentFiles[i] != "")
		{
			inidata += RecentFiles[i];
			inidata += QChar(10);
			astr.setNum(RecentFileType[i]);
			inidata += astr; 
			inidata += QChar(10);
		}
	}
 	QByteArray ba = inidata.toLatin1();
 	const char* buf = ba.data();
 	//write the settings file
	QFile nfile(inifile);
	if(nfile.open(QIODevice::ReadWrite | QIODevice::Truncate))
	{
		nfile.write(buf,ba.size());
		nfile.close();
	}
}

void MainWindowImpl::actionRecentFiles()
{
	QAction* pact = (QAction*)sender();
	//get which recent file was clicked
	int chosen = 0;
	for(;chosen < 8;chosen++)
	{
		if(pact == RecentFilesAction[chosen])
			break;
	}
	fn = pact->data().toString();
	int i = RecentFileType[chosen];
	QMdiSubWindow* wr;
	QString type;
	int j = fn.lastIndexOf("/");
	//get the type of file
	switch(i)
	{
		case PRGFile:
			type = PRGType;
			Prgdir = fn.left(j);
			break;
		case TextFile:
			type = BasicType;
			Txtdir = fn.left(j);
			break;
		case DataFile:
			type = DataType;
			Datadir = fn.left(j);
			break;
		//a d64 file
		case D64File:
			D64Dir = fn.left(j);
			int daal = DoD64File();
			if(daal == 3)
			{
				wr = (QMdiSubWindow*)mdiArea->currentSubWindow();
				MdiChild* child = (MdiChild*)wr->widget();
				child->LastOpenedD64 = fn;
				child->FileOnDisk = false;
			}
			else
			{
				//if the file does not exist any more.
				if(daal == 0)
				{
					RecentFiles[chosen] = "";
					RecentFilesAction[chosen]->setVisible(false);
					statusBar()->showMessage("Error loading file.",10000);
				}
			}
			break;
	}
	//not a d64 file
	if(i != D64File)
	{
		QMdiSubWindow *existing = findMdiChild(fn);
		//if already open just switch to the window containing the file
		if(existing)
			mdiArea->setActiveSubWindow(existing);
		else
		{
			if(SubLoad(fn,type))
			{
				wr = (QMdiSubWindow*)mdiArea->currentSubWindow();
				MdiChild* child = (MdiChild*)wr->widget();
				QString grd = fn.right(fn.length() - j -1);
				grd += "  Loaded";
				statusBar()->showMessage(grd,10000);
				child->FileOnDisk = true;
			}
			else
			{
				//if the file does not exist any more remove from list.
				RecentFiles[chosen] = "";
				RecentFilesAction[chosen]->setVisible(false);
				statusBar()->showMessage("Error loading file.",10000);
			}
		}
	}
}

void MainWindowImpl::on_actionPRGnew_activated()
{
	//create new PRG file
	MdiChild* child = createMdiChild();
	child->newFile(PRGFile);
	//if only one file open just maximize it
	if(mdiArea->subWindowList().count() == 1)
		child->showMaximized();
	else
	{		
		child->showNormal();
		if(arrange == 1)
			on_actionCascade_activated();
		else
			on_actionTile_activated();
	}
	fn = child->curFile;
	//setup the character position display hook
	connect(child,SIGNAL(cursorPositionChanged()),this,
		SLOT(DeCursorPositionChanged()));
	QString grd = fn;
	grd += "  Loaded";
	statusBar()->showMessage(grd,10000);
	//set upper case commodore64 font
	on_actionPetUpperCase_activated();
}

void MainWindowImpl::on_actionTXTnew_activated()
{
	//create new text - asm file
	MdiChild* child = createMdiChild();
	child->newFile(TextFile);
	child->InitLineNumber();
	if(mdiArea->subWindowList().count() == 1)
		child->showMaximized();
	else
	{
		child->showNormal();
		if(arrange == 1)
			on_actionCascade_activated();
		else
			on_actionTile_activated();
	}
	fn = child->curFile;
	//setup the character position display hook
	connect(child,SIGNAL(cursorPositionChanged()),this,
		SLOT(cursorPositionChanged()));
	//setup tab stops and line highlight
	child->setTabStopWidth(Tabspace.toInt() * (Txtfntsz.toInt() - 1));
	if(TxtLineHighlight == "1")
		child->HighlightTxt = true;
	else
		child->HighlightTxt = false;
	QString grd = fn;
	grd += "  Loaded";
	statusBar()->showMessage(grd,10000);
}

void MainWindowImpl::on_actionPetUpperCase_activated()
{
	//Set the upper case font for only PRG file
	QMdiSubWindow* wr = (QMdiSubWindow*)mdiArea->currentSubWindow();
	if(wr)
	{
		MdiChild* child = (MdiChild*)wr->widget();
		if(child->typeFile == PRGFile)
		{
			PRGfont = QFont("PET SCII set 1");
			Prgfnt = "PET SCII set 1";
			PRGfont.setPointSize(Prgfntsz.toInt());
			wr->setFont(PRGfont);
			wr->hide();
			if(mdiArea->subWindowList().count() == 1)
				wr->showMaximized();
			else
				wr->showNormal();
		}
	}
}

void MainWindowImpl::on_actionPetLowerCase_activated()
{
	//Set the lower case font for only PRG file
	QMdiSubWindow* wr = (QMdiSubWindow*)mdiArea->currentSubWindow();
	if(wr)
	{
		MdiChild* child = (MdiChild*)wr->widget();
		if(child->typeFile == PRGFile)
		{
			PRGfont = QFont("PET News set 2");
			Prgfnt = "PET News set 2";
			PRGfont.setPointSize(Prgfntsz.toInt());
			wr->setFont(PRGfont);
			wr->hide();
			if(mdiArea->subWindowList().count() == 1)
				wr->showMaximized();
			else
				wr->showNormal();
		}
	}
}

void MainWindowImpl::on_actionRenumber_activated()
{
	//main renumber routine
	QMdiSubWindow* wr = (QMdiSubWindow*)mdiArea->currentSubWindow();
	if(wr)
	{
		MdiChild* child = (MdiChild*)wr->widget();
		if((child->typeFile == PRGFile) && !RenumberInProgress)
		{
			//only let one instance of renumber at the same time
			RenumberInProgress = true;
			int rnmer = Renumber();
			//was the renumber sucessful?
			if(rnmer != 0)
			{
				QString erstr;
				//print the error message and the line number
				erstr.sprintf("There has been an error while renumbering on line %d.",
					rnmer);
				QMessageBox::critical( this,"Renumber Error!",erstr);
			}
			RenumberInProgress = false;
		}
	}
}

void MainWindowImpl::on_actionCascade_activated()
{
	//cascade the windows
	mdiArea->cascadeSubWindows();
	QSize sz = mdiArea->size();
	foreach(QMdiSubWindow *window, mdiArea->subWindowList())
		window->resize((sz.width()*3)/4,(sz.height()*3)/4);
	arrange = 1;
}

void MainWindowImpl::on_actionTile_activated()
{
	//tile the windows
	mdiArea->tileSubWindows();
	arrange = 2;
}

void MainWindowImpl::on_actionMinimize_activated()
{
	//minimize the active window
	foreach(QMdiSubWindow *window, mdiArea->subWindowList())
		window->showMinimized();
}

void MainWindowImpl::on_actionAbout_BasicEditor_activated()
{
	//about basic editor
	AboutDialog = new QDialog();
	Ui_AboutditorDialog abt;
	abt.setupUi(AboutDialog);
	AboutDialog->show();
	AboutDialog->exec();
	delete AboutDialog;
	AboutDialog = NULL;
}

void MainWindowImpl::on_actionHelp_activated()
{
	//help on editor
	QTextBrowser* m_txtContent = new QTextBrowser();
#ifndef WIN_32
	QString homehlp = homedir+"/CommodoreHelp.html";
#else
	QString homehlp = "CommodoreHelp.html";
#endif
	m_txtContent->setSource(QUrl(homehlp));
	m_txtContent->setWindowTitle("Commodore Basic Editor Help");
	m_txtContent->show();
	m_txtContent->resize(600,400);
}

void MainWindowImpl::on_actionAbout_Qt_activated()
{
	//about Qt
	Aboutqt = new QDialog();
	Ui_AboutQtDialog abtq;
	abtq.setupUi(Aboutqt);
	Aboutqt->show();
	Aboutqt->exec();
	delete Aboutqt;
	Aboutqt = NULL;
}

void MainWindowImpl::on_toolButton_9_clicked()
{
	//help on editor
	on_actionHelp_activated();
}

void MainWindowImpl::on_actionConfiguration_activated()
{
	//configuration dialog
	on_toolButton_6_clicked();
}

void MainWindowImpl::on_actionConfig_activated()
{
	//configuration dialog do the strings - tab setup
	//setup the strings in the dialog from the data read in the settings file
	CfgDialog = new QDialog();
	cfgdlg = new Ui_ConfigDialog();
	cfgdlg->setupUi(CfgDialog);
	cfgdlg->lineEditPRGDir->setText(Prgdir);
	cfgdlg->BasTxtlineEdit->setText(Txtdir);
	cfgdlg->DatalineEdit->setText(Datadir);
	cfgdlg->AssemCmdLn->setText(AsmCmdLine);
	cfgdlg->PRGspinBox->setValue(Prgfntsz.toInt());
	cfgdlg->BasTxtspinBox->setValue(Txtfntsz.toInt());
	cfgdlg->DataspinBox->setValue(Datafntsz.toInt());
	cfgdlg->PRGfontComboBox->setCurrentFont(PRGfont);
	cfgdlg->BasTxtfontComboBox->setCurrentFont(Txtfont);
	cfgdlg->DatafontComboBox->setCurrentFont(Datafont);
	cfgdlg->TabspinBox->setValue(txttab);
	if(TxtLineHighlight == "1")
		cfgdlg->highlightCheckBox->setCheckState(Qt::Checked);
	connect(cfgdlg->PRGDirpushButton,SIGNAL(clicked()),SLOT(actionConfigDirChoise()));
	connect(cfgdlg->BasTxtpushButton,SIGNAL(clicked()),SLOT(actionConfigDirChoise()));
	connect(cfgdlg->DatapushButton,SIGNAL(clicked()),SLOT(actionConfigDirChoise()));
	CfgDialog->setWindowFlags(Qt::WindowStaysOnTopHint);
	CfgDialog->show();
	CfgDialog->exec();
	if(CfgDialog->result())
	{
		//get the data from the dialog and place it in the variables
		Prgdir = cfgdlg->lineEditPRGDir->text();
		Txtdir = cfgdlg->BasTxtlineEdit->text();
		Datadir = cfgdlg->DatalineEdit->text();
		AsmCmdLine = cfgdlg->AssemCmdLn->text();
		int i = cfgdlg->PRGspinBox->value();
		Prgfntsz.setNum(i);
		PRGfont = cfgdlg->PRGfontComboBox->currentFont();
		PRGfont.setPointSize(i);
		Prgfnt = cfgdlg->PRGfontComboBox->currentText();
		Datafont = cfgdlg->DatafontComboBox->currentFont();
		i = cfgdlg->DataspinBox->value();
		Datafntsz.setNum(i);
		Datafont.setPointSize(i);
		Datafnt = cfgdlg->DatafontComboBox->currentText();
		Txtfont = cfgdlg->BasTxtfontComboBox->currentFont();
		i = cfgdlg->BasTxtspinBox->value();
		Txtfntsz.setNum(i);
		Txtfont.setPointSize(i);
		Txtfnt = cfgdlg->BasTxtfontComboBox->currentText();
		txttab = cfgdlg->TabspinBox->value();
		Tabspace.setNum(txttab); 
		QMdiSubWindow* wr = (QMdiSubWindow*)mdiArea->currentSubWindow();
		if(wr)
		{
			MdiChild* child = (MdiChild*)wr->widget();
			//if this is a text - asm file set the tab width
			if(child->typeFile == TextFile)
				child->setTabStopWidth(txttab * (i - 1));
		}
		if(cfgdlg->highlightCheckBox->isChecked())
			TxtLineHighlight = "1";
		else
			TxtLineHighlight = "0";
	}
	delete cfgdlg;
	delete CfgDialog;
	CfgDialog = NULL;
}

void MainWindowImpl::on_toolButton_6_clicked()
{
	//the insert special character dialog - not modal
	QMdiSubWindow* wr = (QMdiSubWindow*)mdiArea->currentSubWindow();
	if(wr)
	{
		MdiChild *mdiChild = qobject_cast<MdiChild *>(wr->widget());
		//if it is a PRG file
		if(mdiChild->typeFile == PRGFile)
		{
			InsDialog = new QDialog();
			Ui_ExtraKeysDialog keyDlg;
			keyDlg.setupUi(InsDialog);
			connect(keyDlg.whiteButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.redButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.greenButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.bluelButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.orangeButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.blackButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.brownButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.ltredButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.ltblueButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.dkgreyButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.ltgreenButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.ltgreyButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.purpleButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.yellowButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.mdgreyButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.cyanButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.uplowButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.curdnButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.revonButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.homeButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.deleteButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.currtButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.returnButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.charupButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.curupButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.revoffButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.clrscrButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.insertButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			connect(keyDlg.curlftButton,SIGNAL(clicked()),SLOT(insertKeysPressed()));
			InsDialog->setWindowFlags(Qt::WindowStaysOnTopHint);
			InsDialog->show();
			InsDialog->exec();
			delete InsDialog;
			InsDialog = NULL;
		}
	}
}

void MainWindowImpl::actionConfigDirChoise()
{
	//get the default diirectories for the different file format
	//coupled on the config dialog
	QPushButton* pb = (QPushButton*)sender();
	//trick - width contains the file type
	int i = pb->iconSize().width();
	QString pst;
	if(i == PRGFile)
		pst = Prgdir;
	else if(i == TextFile)
		pst = Txtdir;
	else
		pst = Datadir;
	QString dir = QFileDialog::getExistingDirectory(this,tr("Set Directory"),
		pst,QFileDialog::ShowDirsOnly | QFileDialog::DontResolveSymlinks);
	if(dir != "")
	{
		if(i == PRGFile)
			cfgdlg->lineEditPRGDir->setText(dir);
		else if(i == TextFile)
			cfgdlg->BasTxtlineEdit->setText(dir);
		else
			cfgdlg->DatalineEdit->setText(dir);
	} 
}

void MainWindowImpl::insertKeysPressed()
{
	//this routine inserts the special characters in a PRG file
	QToolButton* bb = (QToolButton*)sender();
	//trick - width contains the number of the button pressed
	ushort charno = (ushort)bb->iconSize().width();
	if(charno < 256)
	{
		QMdiSubWindow* wr = (QMdiSubWindow*)mdiArea->currentSubWindow();
		MdiChild *mdiChild = qobject_cast<MdiChild *>(wr->widget());
		QTextCursor cursor = mdiChild->textCursor();
		cursor.insertText(QChar(charno));
	}
}

void MainWindowImpl::on_actionFont_activated()
{
	//set the type and size of the font for the different file types
	bool ok;
	QMdiSubWindow* wr = (QMdiSubWindow*)mdiArea->currentSubWindow();
	if(wr != NULL)
	{
		QFont pfont;
		MdiChild *mdiChild = qobject_cast<MdiChild *>(wr->widget());
		if(mdiChild->typeFile == PRGFile)
			pfont = PRGfont;
		else if(mdiChild->typeFile == TextFile)
			pfont = Txtfont;
		else
			pfont = Datafont;
		QFont font = QFontDialog::getFont(&ok,pfont,this);
	   if(ok)
		{
			int fontsize = font.pointSize();
			if(mdiChild->typeFile == PRGFile)
			{
				Prgfntsz.setNum(fontsize);
				PRGfont = font;
			}
			else if(mdiChild->typeFile == TextFile)
			{
				Txtfntsz.setNum(fontsize);
				Txtfont = font;
				//for text - asm files also set the tab width related to the font
				mdiChild->setTabStopWidth(Tabspace.toInt() * (fontsize - 1));
			}
			else
			{
				Datafntsz.setNum(fontsize);
				Datafont = font;
			}
			wr->setFont(font);
			wr->hide();
			if(mdiArea->subWindowList().count() == 1)
				wr->showMaximized();
			else
				wr->showNormal();
		}
	}
}

void MainWindowImpl::on_actionOpen_PRG_File_activated()
{
	//open a PRG file
	OpenPRG(PRGFile);
}

void MainWindowImpl::on_actionOpen_BAS_or_Text_activated()
{
	//open a text - asm file
	OpenPRG(TextFile);
}

void MainWindowImpl::on_actionSave_Data_File_activated()
{
	//Save the data file as another file - not the displayed text
	QMdiSubWindow* wr = (QMdiSubWindow*)mdiArea->currentSubWindow();
	if(wr)
	{
		MdiChild *mdiChild = qobject_cast<MdiChild *>(wr->widget());
		if(mdiChild->typeFile == DataFile)
		{
			QString curfil = mdiChild->userFriendlyCurrentFile();
			QString fileName = QFileDialog::getSaveFileName(this,tr("Save As"),
				curfil + ".cdt");
			if(!fileName.isEmpty())
				mdiChild->saveFile(fileName);
		}
	}
}

void MainWindowImpl::on_actionOpenD64_activated()
{
	//open a d64 file and display the contents for the actions load - rename ...
	OpenPRG(D64File);
}

void MainWindowImpl::D64ButtonPressed()
{
	//get the button that was chicked and exit the dialog
	QPushButton* pb = (QPushButton*)sender();
	//trick - width contains the button pressed
	d64but = pb->iconSize().width();
	QDialog* qd = (QDialog*)pb->parent();
	qd->accept();
}

void MainWindowImpl::on_actionSaveD64_activated()
{
	//save a file to a d64 file
	QMdiSubWindow* wr = (QMdiSubWindow*)mdiArea->currentSubWindow();
	if(wr)
	{
		MdiChild *mdiChild = qobject_cast<MdiChild *>(wr->widget());
		//if this is a new fie
		if(mdiChild->curFile == "")
		{
			fnlog = new QDialog();
			//get a filename to save to d64
			Ui_D64FilenameDialog fdlg;
			fdlg.setupUi(fnlog);
			fnlog->setWindowFlags(Qt::WindowStaysOnTopHint);
			fnlog->show();
			fnlog->exec();
			if(fnlog->result())
			{
				mdiChild->curFile = fdlg.lineEdit->text();
				mdiChild->setWindowTitle(mdiChild->curFile + "[*]");
				mdiChild->SetFile();
				delete fnlog;
				fnlog = NULL;
			}
			else
			{
				delete fnlog;
				fnlog = NULL;
				return;
			}
		}
		//setup a simple message box with extra buttons
		QMessageBox msgBox;
		msgBox.setText("The document has been modified.");
		msgBox.setWindowTitle("Save to D64 File");
		msgBox.setInformativeText("Do you want to save your changes?");
		QPushButton* pbp = msgBox.addButton("SaveTo..",QMessageBox::YesRole);
		msgBox.setStandardButtons(QMessageBox::Save | QMessageBox::Cancel);
		msgBox.setDefaultButton(QMessageBox::Save);
 		QString ofile;
 		int ret = msgBox.exec();
 		//save with name change
 		if (msgBox.clickedButton() == pbp)
			ofile = QFileDialog::getSaveFileName(this,
				tr("Save to D64 file)"),mdiChild->LastOpenedD64,D64Type);
		else
		{
			if(ret == QMessageBox::Save)
			{
				//default saves to the last opened d64 file if exists
				ofile = mdiChild->LastOpenedD64;
				if(ofile == "")
					ofile = QFileDialog::getSaveFileName(this,
						tr("Save to D64 file)"),mdiChild->LastOpenedD64,D64Type);
				if(ofile == "")
					return;
			}
			else
				return;
		}
		//buffer large enough for an extended d64 file
		uchar* d64 = new uchar[226000];
		QFile file(ofile);
		if(!file.open(QFile::ReadWrite))
		{
			QMessageBox::warning(this,tr("Save Error!"),tr("Cannot save file %1:\n%2.")
				.arg(mdiChild->curFile).arg(file.errorString()));
			return;
		}
		int d64sz = file.size();
		file.read((char*)d64,d64sz);
		file.close();
		union	{uchar ts[2];ushort sts;} xleng;
		ushort totSectors = 0;
		int i,j,drpt;
		char cbuf[24];
		QString fi;
		int m = 256;
		int fileNameLength;
		//if the filename contains a directory path then just get the file name
		if(mdiChild->curFile.contains("/"))
		{
			fileNameLength = mdiChild->curFile.length() - mdiChild->curFile.lastIndexOf("/") - 1;
			fi = mdiChild->curFile.right(fileNameLength);
		}
		else
		{
			fi = mdiChild->curFile;
			fileNameLength = fi.length();
		}
		//limit it to the standard max file name length for the commodore
		if(fileNameLength > 16)
			fileNameLength = 16;
		strncpy(cbuf,fi.toLocal8Bit().constData(),fileNameLength);
		uchar nextSector = 1;
		//see if the file already exists on the disk
		do
		{
			drpt = int(nextSector*256) + Track18Sector0;
			nextSector = d64[drpt+1];
			if(d64[drpt] == 0)
				m = nextSector;
			for(i = 2;i < m;)
			{
				j = 0;
				if(d64[drpt+i] & 127)
				{
					xleng.ts[0] = d64[drpt+i+28];
					xleng.ts[1] = d64[drpt+i+29];
					totSectors += xleng.sts;
					for(;j < fileNameLength;j++)
					{
						if(d64[drpt+i+j+3] != cbuf[j])
							break;
					}
					if(j == fileNameLength)
					{
						//if the file name matches also with length 0xa0 is term byte
						if(fileNameLength == 16)
							break;
						if(d64[drpt+i+j+4] == 0xa0)
							break;
					}
				}
				i+=32;
			}
			if(j == fileNameLength)
				break;
		}while(d64[drpt] != 0);
		int splines = mdiChild->document()->lineCount() - 1;
		//if the last line contains a space then give an extra line
		if(mdiChild->document()->findBlockByNumber(splines).text().contains(QChar(' ')))
			splines++;
		//if a PRG file has no lines or has been modified then translate it into
		//a real commodore tokenized basic file
		if(((mdiChild->lines == NULL) || mdiChild->document()->isModified()) &&
			((mdiChild->typeFile == PRGFile)))
		{
			if(!PreSave(mdiChild,splines))
			{
				//tokenizing error occured
				QMessageBox::critical(this,tr("Analyzing Error! Can not continue!"),tr("There has been an error constructing the file.\nCheck for syntax and other errors."));
				delete d64;
				return;
			}
		}
		uchar* savbuf;
		int ki;
		if(mdiChild->typeFile == PRGFile)
		{
			savbuf = new uchar[88 * splines];
			ki = PostSave(mdiChild,savbuf,splines);
		}
		else
		{
			if(mdiChild->typeFile == DataFile)
			{
				savbuf = new uchar[mdiChild->dataBuffersz + 16];
				for(ki = 0;ki < mdiChild->dataBuffersz;ki++)
					savbuf[ki] = mdiChild->dataBuffer[ki];
			}			
			else
			{
				//normal text file save to d64
				int k = mdiChild->document()->blockCount();
				ki = mdiChild->document()->characterCount() + k - 2;
				savbuf = new uchar[ki + 256];
				savbuf[0] = 0;
				for(m = 0;m < k;m++)
				{
					strcat((char*)savbuf,
						mdiChild->document()->findBlockByNumber(m).text().toLocal8Bit());
					strcat((char*)savbuf,"\n");
				}
			}
		}
		//if the file already exists on the disk
		if(j == fileNameLength)
		{
			//start by finding the starting track and sector
			union	{uchar ts[4];int sts;} fleng;
			fleng.ts[0] = d64[drpt+i+28];
			fleng.ts[1] = d64[drpt+i+29];
			fleng.ts[2] = 0;
			fleng.ts[3] = 0;
			int tzok = calcOff(d64[drpt+i+1],d64[drpt+i+2]);
			int p = 0;
			m = 0;
			int n;
			char* tracks = new char[fleng.sts+2];
			char* sectors = new char[fleng.sts+2];
			tracks[0] = d64[drpt+i+1];
			sectors[0] = d64[drpt+i+2];
			//write over the first sector - this is so far only in memory
			do
			{
				for(n = 2;n < 256;n++)
				{
					d64[tzok+n] = savbuf[p++];
					if(p == ki)
					{
						d64[tzok] = 0;
						d64[tzok+1] = n;
						break;
					}
				}
				tzok = calcOff(d64[tzok],d64[tzok+1]);
				m++;
				tracks[m] = d64[tzok];
				sectors[m] = d64[tzok+1];
			}while((p != ki) && (m < fleng.sts));
			union	{uchar ts[4];int sts;} flend;
			int drlt = Track18Sector0;	//four bytes before begining of bam. tracks start at 1
			char ntrack;		//so this makes up for the offset
			char nsector;
			bool found;
			//if the same number of sectors as the orignal file
			if(p == ki)
			{
				found = true;
				if(fleng.sts > m)
				{
					ntrack = tracks[m-1];
					nsector = sectors[m-1];
					flend.sts = m;
					d64[drpt+i+28] = flend.ts[0];
					d64[drpt+i+29] = flend.ts[1];
					while(ntrack != 0)
					{
						i = (int)ntrack<<2;
						d64[drlt+i]++;
						i += (int)((nsector>>3) + 1);
						d64[drlt+i] |= (1<<(nsector & 7));
						tzok = calcOff(ntrack,nsector);
						ntrack = d64[tzok];
						nsector = d64[tzok+1];
					}
				}
			}
			else
			{
				j = ki/254;
				if((j * 254) != ki)
					j++;
				flend.sts = j;
				//new file length set
				d64[drpt+i+28] = flend.ts[0];
				d64[drpt+i+29] = flend.ts[1];
				//write to memory, return success if it fits on the disk
				found = WriteD64(tracks[m-2],sectors[m-2],p,ki,d64,savbuf);
			}
			delete sectors;
			delete tracks;
			//if sucessful the write the memory to disk
			if(found)
			{
				if(file.open(QIODevice::ReadWrite | QIODevice::Truncate))
				{
					file.write((char*)d64,d64sz);
					file.close();
					mdiChild->document()->setModified(false);
					mdiChild->setWindowModified(false);
				}
			}
			else
				QMessageBox::critical(this,tr("Disk Full!"),tr("The disk does not have enough free space to save this file."));
		}
		else
		{
			//new file to place on the disk
			int s = ki/254;
			if((s * 254) != ki)
				s++;
			if((totSectors + s) > 664)
			{
				QMessageBox::critical(this,tr("Disk Full!"),tr("The disk does not have enough free space to save this file."));
				delete savbuf;
				delete d64;
				return;
			}
			drpt = Track18Sector1;
			//bam entry for directory sectors
			int drxt = BeginDirBam;
			bool fnd = false;
			//find an open directory entry
			while((drpt < EndDirBam) && !fnd)
			{
				for(i = 2;i < 256;i+=32)
				{
					if((d64[drpt+i] & 127) == 0)
					{
						fnd = true;
						break;
					}
				}
				if(!fnd)
				{
					if(d64[drpt] != 0)
						drpt = (d64[drpt+1]*256) + Track18Sector0;
					else
					{
						if(d64[drxt] == 0)
							break;
						else
						{
							d64[drxt]--;
							for(m = 1;m < 4;m++)
							{
								if(d64[drxt+m] != 0)
									break;
							}
							uchar rsector = (char)(m-1)<<3;
							uchar r = 1;
							int t = 0;
							for(;t < 8;t++)
							{
								if(r & d64[drxt+m])
									break;
								r = r<<1;
							}
							rsector += t;
							r ^= 255;
							d64[drxt+m] = d64[drxt+m] & r;
							if(rsector > 18)
								break;
							d64[drpt] = 18;
							d64[drpt+1] = rsector;
							drpt = (rsector*256) + Track18Sector0;
							d64[drpt] = 0;
							d64[drpt+1] = 255;
						}
					} 
				}
			}
			if(!fnd)
				QMessageBox::critical(this,tr("Directory Full!"),tr("There is no free directory entyr, can not save file."));
			else
			{
				drpt += i;
				i = 0;
				//fill in the file name
				for(m = 3;i < fileNameLength;m++)
					d64[drpt+m] = cbuf[i++];
				//fill the rest to 16 characters with shift space character
				for(;m < 19;m++)
					d64[drpt+m] = 160;
				xleng.sts = s;
				//file length in bytes
				d64[drpt+28] = xleng.ts[0];
				d64[drpt+29] = xleng.ts[1];
				int tzon = Track18Sector0;
				uchar ntrack,nsector;
				//offset to track 17 in bam entry
				i = 68;
				//offset to track 19 in bam entry
				j = 76;
				//start searching for free sectors both directoions 0 - 17 19 - 35
				for(m = 0;m < 17;m++)
				{
					if(d64[tzon+j] != 0)
						break;
					if(d64[tzon+i] != 0)
						break;
					j+=4;
					i-=4;
				}
				uchar rx = 1;
				int tx = 0;
				if(d64[tzon+i] != 0)
				{
					ntrack = (uchar)i>>2;
					for(m = 1;m < 4;m++)
					{
						if(d64[tzon+i+m] != 0)
						{
							nsector = (m-1)<<3;
							for(;tx < 8;tx++)
							{
								if(rx & d64[tzon+i+m])
									break;
								rx = rx<<1;
							}
							rx = rx^255;
							d64[tzon+i+m] = d64[tzon+i+m] & rx;
							nsector += tx;
							break;
						}
					}
					d64[tzon+i]--;
				}
				else if(d64[tzon+j] != 0)
				{
					ntrack = (uchar)j>>2;
					for(m = 1;m < 4;m++)
					{
						if(d64[tzon+j+m] != 0)
						{
							nsector = (m-1)<<3;
							for(;tx < 8;tx++)
							{
								if(rx & d64[tzon+j+m])
									break;
								rx = rx<<1;
							}
							rx ^= 255;
							d64[tzon+j+m] = d64[tzon+j+m] & rx;
							nsector += tx;
							break;
						}
					}
					d64[tzon+j]--;
				}
				d64[drpt] = 130;
				d64[drpt+1] = ntrack;
				d64[drpt+2] = nsector;
				drpt = calcOff(ntrack,nsector);
				fnd = true;
				//write the first sector
				for(m = 2,drxt = 0;m < 256;m++,drxt++)
				{
					d64[drpt+m] = savbuf[drxt];
					if(drxt == ki)
						break;					
				}
				//if there is more to write
				if(drxt != ki)
					fnd = WriteD64(ntrack,nsector,drxt,ki,d64,savbuf);
				//write the memory to disk
				if(fnd)
				{
					if(file.open(QIODevice::ReadWrite | QIODevice::Truncate))
					{
						file.write((char*)d64,d64sz);
						file.close();
						mdiChild->document()->setModified(false);
						mdiChild->setWindowModified(false);
					}
				}
				else
					QMessageBox::critical(this,tr("Disk Full!"),tr("The disk does not have enough free space to save this file."));
			}
		}
		delete savbuf;
		delete d64;
	}
}

bool MainWindowImpl::WriteD64(uchar ntrack,uchar nsector,int p,int ki,uchar* d64,
	uchar* savbuf)
{
	bool found = false;
	int drlt = Track18Sector0;
	int k,m,j;
	bool recnt = false;
	bool cntdir;
	//whitch direction to begin the search
	if(ntrack > 18)
		cntdir = true;
	else
		cntdir = false;
	uchar xtrak,xsect,shft;
	while(p < ki)
	{
		int i = (int)ntrack<<2;	//the pointer to the entry in the bam start
		if(cntdir)					//here and go up for search
		{	//find another free entry on the track but not a directory sector
			for(;((d64[drlt+i] == 0) && (i < 144)) || (i == 72);i+=4);
			if(i > 142)			//if beyond the end then break and
			{						//set the flag to search backwards
				if(recnt)		//if already set then the disk is full
					break;
				recnt = true;
				cntdir = false;
				i = 140;
				found = false;
			}
			else
			{
				for(k = 1;k < 4;k++)
				{
					if(d64[drlt+i+k] != 0)
						break;
				}
				xtrak = i>>2;
				xsect = (k-1)<<3;
				for(shft = 1,m = 0;m < 8;m++)
				{
					if(d64[drlt+i+k] & shft)
					{
						xsect += m;
						break;
					}
					shft = shft<<1;
				}
				found = true;
			}
		}
		else
		{
			for(;(d64[drlt+i] == 0) || (i == 72);i-=4);
			if(i != 0)
			{
				for(k = 1;k < 4;k++)
				{
					if(d64[drlt+i+k] != 0)
						break;
				}
				xtrak = i>>2;
				xsect = (k-1)<<3;
				for(shft = 1,m = 0;m < 8;m++)
				{
					if(d64[drlt+i+k] & shft)
					{
						xsect += m;
						break;
					}
					shft = shft<<1;
				}
				found = true;
			}
			else
			{
				if(recnt)
					break;
				recnt = true;
				cntdir = true;
				found = false;
			}
		}
		//if enough free sectors are found
		if(found)
		{
			shft ^= 255;
			d64[drlt+i]--;
			d64[drlt+i+k] &= shft;
			j = calcOff(ntrack,nsector);
			d64[j] = xtrak;
			d64[j+1] = xsect;
			j = calcOff(xtrak,xsect);
			ntrack = xtrak;
			nsector = xsect;
			for(i = 2;i < 256;i++)
			{
				d64[j+i] = savbuf[p++];
				if(p == ki)
				{
					d64[j] = 0;
					d64[j+1] = (uchar)i & 255;
					break;
				}
			}
		}
	}
	return found;
}

int MainWindowImpl::calcOff(uchar track,uchar sector)
{
	//get the offset in the buffer for the given track and sector
	int drpt;
	if(track < 19)
		drpt = (track-1)*5376;
	else if((track > 18) && (track < 26))
		drpt = Track18Sector0 + (track-18)*4864;
	else if((track > 25) && (track < 32))
		drpt = 125440 + (track-25)*4608;
	else
		drpt = 153088 + (track-31)*4352;
	drpt += (sector*256);
	return drpt;	
}

void MainWindowImpl::on_actionOpen_Data_File_activated()
{
	OpenPRG(DataFile);
}

void MainWindowImpl::on_actionSave_as_PRG_activated()
{
	if(!SavePRGFile())
		QMessageBox::critical( this, "Save File Error!",
			"There has been an error while saving the file.");
}

void MainWindowImpl::on_actionSave_as_BAS_Text_activated()
{
	//save a text file
	QMdiSubWindow* wr = (QMdiSubWindow*)mdiArea->currentSubWindow();
	if(wr)
	{
		MdiChild *mdiChild = qobject_cast<MdiChild *>(wr->widget());
		QString fileName = QFileDialog::getSaveFileName(this,
			tr("Save Text file)"),mdiChild->curFile
			,"All Files (*);;ASM Files (*.ASM *.asm);;Text Files (*.TXT txt)");
		if(!fileName.isEmpty())
			mdiChild->saveFile(fileName);
	}
}

void MainWindowImpl::on_toolButton_clicked()
{
	OpenPRG(PRGFile);
}

void MainWindowImpl::on_toolButton_2_clicked()
{
	on_actionSave_as_PRG_activated();
}

void MainWindowImpl::on_toolButton_5_clicked()
{
	on_actionRenumber_activated();
}

void MainWindowImpl::on_toolButton_7_clicked()
{
#ifndef QT_NO_PRINTER
	QMdiSubWindow* m = (QMdiSubWindow*)mdiArea->currentSubWindow();
	if (m)
	{
		MdiChild* mc = (MdiChild*)m->widget();
		mc->print();		
	}
	else
		statusBar()->showMessage("No File Loaded",10000);
#endif
}

void MainWindowImpl::on_toolButton_8_clicked()
{
	//find a string in the text then goto the line
	QMdiSubWindow *window = mdiArea->activeSubWindow();
	if(window)
	{
		MdiChild *mdiChild = qobject_cast<MdiChild *>(window->widget());
		QTextCursor cur = mdiChild->textCursor();
		QString str = cur.selection().toPlainText();
		FndDlg = new QDialog();
		Ui_FindDialog fnd;
		fnd.setupUi(FndDlg);
		if(str == "")
			fnd.lineEdit->setText(fndtxt);
		else
			fnd.lineEdit->setText(str);
		fnd.lineEdit->setFocus();
		if(prst & 2)
			fnd.checkBox->setCheckState(Qt::Checked);
		if(prst & 4)
			fnd.checkBox_2->setCheckState(Qt::Checked);
		FndDlg->setWindowFlags(Qt::WindowStaysOnTopHint);
		FndDlg->show();
		FndDlg->exec();
		if(FndDlg->result())
		{
			fndtxt = fnd.lineEdit->text();
			if(fndtxt != "")
			{
				prst = (int)fnd.checkBox->isChecked() << 1;
				prst += (int)fnd.checkBox_2->isChecked() << 2;
				prst += (int)fnd.checkBox_3->isChecked();
				if(!mdiChild->find(fndtxt,(QTextDocument::FindFlags)prst))
					statusBar()->showMessage("Text not found",10000);
			}
		}
		delete FndDlg;
		FndDlg = NULL;
	}
}

void MainWindowImpl::on_assemble_button_clicked()
{
	//assemble a assembler source file with the chosen assembler
	QMdiSubWindow* wr = (QMdiSubWindow*)mdiArea->currentSubWindow();
	if(wr != NULL)
	{
		MdiChild* child = (MdiChild*)wr->widget();
		if(child->typeFile == TextFile)
		{
			QStringList fls;
			fls << child->curFile;
			//first save the file being edited
			child->saveFile(child->curFile);
			QProcess exeasm;
			QString ren;
			//setup the dialog to receive the output of the assembler
			AsmResDialog = new QDialog();
			Ui_AsmResult arDlg;
			arDlg.setupUi(AsmResDialog);
			if(exeasm.execute(AsmCmdLine,fls))
				ren = "Error occured in assembling!";
			else
				ren = "   Assembly successful.";
			//set the results of the action as a string
			arDlg.asmrslt->setText(ren);
			//the output of the assembler is always a text file called err.txt
			QFile erfile("err.txt");
			if(erfile.open(QIODevice::ReadOnly))
			{
				int i = erfile.size();
				char* erbuf = new char[i + 2];
				erfile.read(erbuf,i);
				erbuf[i] = 0;
				erfile.close();
				QString errdata(erbuf);
				arDlg.AsmRslt->setText(errdata);
			}
			AsmResDialog->setWindowFlags(Qt::WindowStaysOnTopHint);
			AsmResDialog->show();
			AsmResDialog->exec();
			delete AsmResDialog;
			AsmResDialog = NULL;
		}			
	}
}

void MainWindowImpl::on_actionAssemble_activated()
{
	on_assemble_button_clicked();
}

void MainWindowImpl::on_actionGoto_Line_activated()
{
	//goto line number for text or to a basic line number
	QMdiSubWindow *window = mdiArea->activeSubWindow();
	if(window)
	{
		MdiChild *mdiChild = qobject_cast<MdiChild *>(window->widget());
		GotoDlg = new QDialog();
		Ui_GotoLineDialog goLine;
		goLine.setupUi(GotoDlg);
		goLine.lineEdit->setFocus();
		GotoDlg->setWindowFlags(Qt::WindowStaysOnTopHint);
		GotoDlg->show();
		GotoDlg->exec();
		if(GotoDlg->result())
		{
			QString gototxt = goLine.lineEdit->text();
			if(gototxt != "")
			{
				int i,k;
				int j = -1;
				//if a basic file goto a basic line
				if(mdiChild->typeFile == PRGFile)
				{
					ushort q = gototxt.toUShort();
					j = mdiChild->document()->lineCount() - 1;
					for(i = 0;i < j;i++)
					{
						if(q <= mdiChild->lines[i].OrignalLineNumber)
							break;
					}
					k = Prgfntsz.toUInt();
				}
				else
				{
					//normal text file
					i = gototxt.toInt() - 1;
					k = Txtfntsz.toInt();
				}
				//if in a basic file the exact line number is not found then
				//goto the closest line
				if(i !=  j)
				{
					//also try to center the line
					QRect qr = mdiChild->geometry();
					int addr = (qr.height() - 150)/(k*2);
					QTextCursor cursor = mdiChild->textCursor();
					int cpos = cursor.blockNumber();
					if(i > cpos)
					{
						cursor.movePosition(QTextCursor::Down,
							QTextCursor::MoveAnchor,i - cpos + addr);
						mdiChild->setTextCursor(cursor);
						cursor.movePosition(QTextCursor::Up,
							QTextCursor::MoveAnchor,addr);
					}
					else
					{
						cursor.movePosition(QTextCursor::Up,
							QTextCursor::MoveAnchor,cpos - i + addr);
						mdiChild->setTextCursor(cursor);
						cursor.movePosition(QTextCursor::Down,
							QTextCursor::MoveAnchor,addr);
					}
					mdiChild->setTextCursor(cursor);
				}
			}
		}
		delete GotoDlg;
		GotoDlg = NULL;
	}
}

void MainWindowImpl::on_lineEdit_textChanged(QString hexStr)
{
	//these two are linked so when one is changed the other changes
	//the hex converter window.
	if(hexStr != h1)
	{
		h1 = hexStr;
		bool ok;
		unsigned int i = hexStr.toUInt(&ok,16);
		//the two are linked so allow only one at a time
		if(ok)
		{
			QString decstr;
			decstr.sprintf("%d",i);
			lineEdit_2->setText(decstr);
		}
	}
}

void MainWindowImpl::on_lineEdit_2_textChanged(QString decStr)
{
	//the decimal converter window
	if(decStr != d1)
	{
		d1 = decStr;
		bool ok;
		unsigned int i = decStr.toUInt(&ok,10);
		if(ok)
		{
			QString hecstr;
			hecstr.sprintf("%X",i);
			lineEdit->setText(hecstr);
		}
	}
}

bool MainWindowImpl::SavePRGFile()
{
	QMdiSubWindow *window = mdiArea->activeSubWindow();
	bool status = false;
	if(window)
	{
		MdiChild *mdiChild = qobject_cast<MdiChild *>(window->widget());
		if(mdiChild->typeFile == PRGFile)
		{
			int splines = mdiChild->document()->lineCount() - 1;
			//if there is an extra char on the last line do one extra I don't know why
			if(mdiChild->document()->findBlockByNumber(splines).text().contains(QChar(' ')))
				splines++;
			if((mdiChild->lines == NULL) || mdiChild->document()->isModified())
			{
				if(!PreSave(mdiChild,splines))
					return false;
			}
			QString an = QFileDialog::getSaveFileName(this,
				tr("Save PRG file)"),mdiChild->curFile,PRGType);
			if (!an.isEmpty())
			{
				//to make sure there is enough space do a little extra 80 is max
				uchar* savbuf = new uchar[100 * splines];
				int k = PostSave(mdiChild,savbuf,splines);
				QFile outfile(an);
				if(outfile.open(QIODevice::ReadWrite) | QIODevice::Truncate)
				{
					outfile.write((const char*)savbuf,k);
					outfile.close();
					status = true;
					mdiChild->setCurrentFile(an);
					fn = an;
				}
				else
					status = false;
				delete savbuf;
			}
		}
	}
	return status;
}

bool MainWindowImpl::PreSave(MdiChild* mdiChild,int splines)
{
	int i;
	ushort* linerel = new ushort [splines];
	QString lbuf;
	for(i = 0;i < splines;i++)
	{
		lbuf = mdiChild->document()->findBlockByNumber(i).text();
		linerel[i] = lbuf.left(lbuf.indexOf(" ")+1).toUShort();			
	}
	PRGLines* mlines =  new PRGLines[i+10];
	mdiChild->ResizePRGLines(i+10);
	ushort ofst = 2049;
	ushort pl;
	for(i = 0;i < splines;i++)
	{
		mlines[i].OrignalLineNumber = linerel[i];
		lbuf = mdiChild->document()->findBlockByNumber(i).text();
		mlines[i].TranslatedLine = lbuf.right(lbuf.length()-lbuf.indexOf(" ")-1);
		pl= MakeLocal(i,0,NULL,mlines,0);
		//error because used later with+ + 2049 > 16bit
		if(pl == 65535)
		{
			QString erstr;
			erstr.sprintf("There has been an error while saving on line %d.",
				linerel[i]);
			QMessageBox::critical( this,"Save Error!",erstr);
			delete[] mlines;
			delete linerel;
			return false;
		}
		ofst += pl;
		mdiChild->lines[i].OrignalLineOffset = ofst;
	}
	for(i = 0;i < splines;i++)
	{
		strcpy((char*)mdiChild->lines[i].OrignalLine,
			(char*)mlines[i].OrignalLine);
		mlines[i].OrignalLineOffset = mdiChild->lines[i].OrignalLineOffset;
		mdiChild->lines[i].TranslatedLine = mlines[i].TranslatedLine;
		mdiChild->lines[i].OrignalLineNumber = mlines[i].OrignalLineNumber;
	}
	delete[] mlines;
	delete linerel;
	return true;
}

int MainWindowImpl::PostSave(MdiChild* mdiChild,uchar* savbuf,int splines)
{
	int k = 2;
	savbuf[0] = 1;
	savbuf[1] = 8;
	//build up the buffer from the basic text
	for(int i = 0;i < splines;i++)
	{
		savbuf[k++] = mdiChild->lines[i].OrignalLineOffset & 255;
		savbuf[k++] = (mdiChild->lines[i].OrignalLineOffset >> 8) & 255;
		savbuf[k++] = mdiChild->lines[i].OrignalLineNumber & 255;
		savbuf[k++] = (mdiChild->lines[i].OrignalLineNumber >> 8) & 255;
		//this can stay at 80 - max line length of encoded basic
		for(int j = 0;j < 80;j++)
		{
			savbuf[k++] = mdiChild->lines[i].OrignalLine[j];
			if(savbuf[k-1] == 0)
				break;
		}
	}
	savbuf[k++] = 0;
	savbuf[k++] = 0;
	return k;
}

bool MainWindowImpl::OpenPRG(int buffer)
{
	//handels all file type loades
	QString header;
	QString type;
	QString dirdir;
	//setup strings and default directory for open file name dialog
	switch(buffer)
	{
		case PRGFile:
			header = PRGHeader;
			type = PRGType;
			dirdir = Prgdir;
			break;
		case TextFile:
			header = BasicHeader;
			type = BasicType;
			dirdir = Txtdir;
			break;
		case DataFile:
			header = DataHeader;
			type = DataType;
			dirdir = Datadir;
			break;
		case D64File:
			header = D64Header;
			type = D64Type;
			dirdir = D64Dir;
			break;
	}
	fn = QFileDialog::getOpenFileName(this,header,dirdir,type,
		//bug in QT of KDE? DontUseNativeDialog needed in getOpenFileName
		NULL,QFileDialog::DontUseNativeDialog);
	if (!fn.isEmpty())
	{
		QMdiSubWindow *existing = findMdiChild(fn);
		int j = fn.lastIndexOf("/");
		//update last viewed directory for different file types
		switch(buffer)
		{
			case PRGFile:
				Prgdir = fn.left(j);
				break;
			case TextFile:
				Txtdir = fn.left(j);
				break;
			case DataFile:
				Datadir = fn.left(j);
				break;
			case D64File:
				D64Dir = fn.left(j);
				break;
		}
		//if the fileis already visible - open
		if(existing)
			mdiArea->setActiveSubWindow(existing);
		else
		{
			int i = 0;
			for(;i < 8;i++)
			{
				if(fn == RecentFiles[i])
					break;
			}
			//if not already in recent file list then add it
			if(i == 8)
			{
				for(i = 0;i < 8;i++)
				{
					if(RecentFiles[i] == "")
					{
						RecentFiles[i] = fn;
						RecentFileType[i] = buffer;
						RecentFilesAction[i] = new QAction(this);
						RecentFilesAction[i]->setData(RecentFiles[i]);
						RecentFilesAction[i]->setVisible(true);
 						connect(RecentFilesAction[i],SIGNAL(triggered()),
 							this,SLOT(actionRecentFiles()));
 						menuRecent_Files->addAction(RecentFilesAction[i]);
						break;
					}
				}
				//if the list is full delete the oldest then add shift add to end
				if(i == 8)
				{
					for(i = 6;i >= 0;i--)
					{
						RecentFiles[i+1] = RecentFiles[i];
						RecentFileType[i+1] = RecentFileType[i];
					}
					RecentFiles[0] = fn;
					RecentFileType[0] = buffer;
				}
			}
			for(i = 0;i < 8;i++)
			{
				if(RecentFiles[i] != "")
				{
					//update the menu labels for recent files.
					int m = RecentFiles[i].lastIndexOf("/") + 1;
					RecentFilesAction[i]->setText(RecentFiles[i].
						right(RecentFiles[i].length()-m));
					RecentFilesAction[i]->setData(RecentFiles[i]);
				}
			}
			if(buffer != D64File)
			{
				//load normal file type
				if(SubLoad(fn,type))
				{
					QMdiSubWindow* wr = (QMdiSubWindow*)mdiArea->currentSubWindow();
					MdiChild* child = (MdiChild*)wr->widget();
					QString grd = fn.right(fn.length() - j -1);
					grd += "  Loaded";
					statusBar()->showMessage(grd,10000);
					child->FileOnDisk = true;
				}
				else
					statusBar()->showMessage("Error loading file.",10000);
			}
			else
			{
				int doi = DoD64File();
				QMdiSubWindow* wr = (QMdiSubWindow*)mdiArea->currentSubWindow();
				if(doi == 3)
				{
					MdiChild* child = (MdiChild*)wr->widget();
					child->LastOpenedD64 = fn;
					child->FileOnDisk = false;
				}
				else
				{
					//if doi !=3 then an error
					wr->deleteLater();
					return false;
				}
				return true;
			}
		}
	}
	else
		return false;
	return true;
}

int MainWindowImpl::DoD64File()
{
	int retwaarde = 3;
	d64Dialog = new QDialog();
	Ui_d64Directory abt64;
	abt64.setupUi(d64Dialog);
	uchar* d64 = new uchar[225024];
	QFile file(fn);
	if(!file.open(QFile::ReadOnly))
	{
		QMessageBox::warning(this,tr("File Not Found"),tr("Cannot read file %1:\n%2.")
			.arg(fn).arg(file.errorString()));
		//error *.d64 file not readable or not found
		return 0;
	}
	int d64sz = file.size();
	file.read((char*)d64,d64sz);
	file.close();
	uchar dskname[24];
	int drpt = DiskLabel;	 //the disk label and id - 91392 begin of track 18
	int j = 0;
	for(;j < 18;j++)
		dskname[j] = d64[drpt+j];
	dskname[j] = 32;
	dskname[j+1] = '-';
	dskname[j+2] = 32;
	dskname[j+3] = d64[drpt+j+3];
	dskname[j+4] = d64[drpt+j+4];
	dskname[j+5] = 0;
	abt64.label->setText((char*)dskname);
	QStringList sl;
	abt64.tableWidget->setColumnCount(4);
	sl << "   Program Name   " << "O" << "Type" << "Blocks";
	abt64.tableWidget->setHorizontalHeaderLabels(sl);
	abt64.tableWidget->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
	abt64.tableWidget->verticalHeader()->hide();
	abt64.tableWidget->setSelectionBehavior(QAbstractItemView::SelectRows);
	int m,n;
	//free sectors on a disk
	int allocated = 663;
	int k = 0;
	int p = 0;
	QTableWidgetItem* qn[8*18*4];
	uchar bgntrack[8*18];
	uchar bgnsector[8*18];
	uchar fileType;
	//first sector of directory
	uchar nextSector = 1;
	uchar nextTrack = 18;
	while(nextTrack != 0)
	{
		drpt = int(nextSector*256) + Track18Sector0;
		nextTrack = d64[drpt];
		nextSector = d64[drpt+1];
		if(nextTrack != 0)
			n = 256;
		else
			n = nextSector;
		union	{uchar ts[2];ushort sts;} fleng;
		for(j = 2;j < n;)
		{
			fileType = d64[drpt+j];
			if(fileType != 0)
			{
				bgntrack[k] = d64[drpt+j+1]; 
				bgnsector[k] = d64[drpt+j+2]; 
				abt64.tableWidget->insertRow(k);
				DirNamePtr[k] = (char*)d64+drpt+j+3;
				//filename
				for(m = 0;m < 16;m++)
					dskname[m] = d64[drpt+j+3+m];
				dskname[m++] = 0;
				QString dnam((char*)dskname);
				qn[p] = new QTableWidgetItem();
				qn[p]->setData(Qt::DisplayRole,dnam);
				abt64.tableWidget->setItem(k,0,qn[p++]);
				qn[p] = new QTableWidgetItem();
				if(!fileType & 0x80)
				{
					dnam = "*";
					qn[p]->setData(Qt::DisplayRole,dnam);
					abt64.tableWidget->setItem(k,1,qn[p]);
				}
				p++;
				switch(fileType & 127)
				{
					//commodore64 file types
					case 0:
						dnam = "DEL";
						break;
					case 1:
						dnam = "SEQ";
						break;
					case 2:
						dnam = "PRG";
						break;
					case 3:
						dnam = "USR";
						break;
					case 4:
						dnam = "REL";
						break;
					default:
						dnam = "???";
						break;
				}
				qn[p] = new QTableWidgetItem();
				qn[p]->setData(Qt::DisplayRole,dnam);
				abt64.tableWidget->setItem(k,2,qn[p++]);
				//file length in two bytes
				fleng.ts[0] = d64[drpt+j+28];
				fleng.ts[1] = d64[drpt+j+29];
				dnam.setNum(fleng.sts);
				//sub this from the total number of sectors
				allocated -= fleng.sts;
				qn[p] = new QTableWidgetItem();
				qn[p]->setData(Qt::DisplayRole,dnam);
				abt64.tableWidget->setItem(k++,3,qn[p++]);
			}
			j+=32;
		}
	}
	QString ss;
	ss.setNum(allocated);
	abt64.labelFreeSectors->setText(ss);
	connect(abt64.pushButton,SIGNAL(clicked()),SLOT(D64ButtonPressed()));
	connect(abt64.pushButton_2,SIGNAL(clicked()),SLOT(D64ButtonPressed()));
	connect(abt64.pushButton_3,SIGNAL(clicked()),SLOT(D64ButtonPressed()));
	connect(abt64.tableWidget,SIGNAL(itemChanged(QTableWidgetItem*)),
		SLOT(fileRename(QTableWidgetItem*)));
	FileNameChanged = false;
	d64Dialog->setWindowFlags(Qt::WindowStaysOnTopHint);
	d64Dialog->show();
	d64Dialog->exec();
	if(d64Dialog->result())
	{
		if(FileNameChanged)
		{
			if(!file.open(QFile::ReadWrite | QIODevice::Truncate))
			{
				QMessageBox::warning(this,tr("Write Error to D64"),
					tr("Cannot update directory entry. %1:\n%2.")
					.arg(fn).arg(file.errorString()));
				return 1;
			}
			file.write((char*)d64,d64sz);
			file.close();
		}
		for(j = 0;j < p;j+=4)
		{
			if(qn[j]->isSelected())
				break;
		}
		if(j < p)
		{
			QMdiSubWindow* wr;
			int q = j>>2;
			//offset in d64 file
			drpt = calcOff(bgntrack[q],bgnsector[q]);
			m = qn[j+3]->text().toInt();
			char* dbuf = new char[(m*256) + 16];
			int rp = 0;
			for(n = 0;n < (m-1);n++)
			{
				for(q = 2;q < 256;q++)
					dbuf[rp++] = d64[drpt+q];
				drpt = calcOff(d64[drpt],d64[drpt+1]);
			}
			for(q = 2;q < (int)d64[drpt+1] + 1;q++)
				dbuf[rp++] = d64[drpt+q];
			dbuf[rp] = 0;
			MdiChild* child = createMdiChild();
			if(d64but == 3)
			{
				child->dataBuffer = new char[rp+16];
				child->dataBuffersz = rp;
				for(n = 0;n < rp;n++)
					child->dataBuffer[n] = dbuf[n];
			}
			child->D64Action = actionSaveD64;
			QString fnam = qn[j]->text().toLatin1();
			for(n = fnam.length() - 1;fnam.at(n) == QChar(0x20) ||
				fnam.at(n) == QChar(0xa0);n--)
			{
				if(n == 1)
					break;
			}
			child->curFile = fnam.left(n+1);
			child->newFile(d64but + 127);
			if(d64but == 1)
			{
				if(child->newPRGLoad(dbuf,rp))
				{
					SetupChild(child);
					child->SetFile();
					wr = (QMdiSubWindow*)mdiArea->currentSubWindow();
					PRGfont.setPointSize(Prgfntsz.toInt());
					wr->setFont(PRGfont);
				}
			}
			else if(d64but == 2)
			{
				child->InitLineNumber();
				child->setTabStopWidth(Tabspace.toInt()*(Txtfntsz.toInt() - 1));
				child->setPlainText(dbuf);
				SetupChild(child);
				child->lastLoad();
				if(TxtLineHighlight == "1")
					child->HighlightTxt = true;
				else
					child->HighlightTxt = false;
				wr = (QMdiSubWindow*)mdiArea->currentSubWindow();
				Txtfont.setPointSize(Txtfntsz.toInt());
				wr->setFont(Txtfont);
			}
			else if(d64but == 3)
			{
				child->SubloadData(dbuf,rp);
				SetupChild(child);
				child->setWindowTitle(child->curFile + "[*]");
				child->SetFile();
				wr = (QMdiSubWindow*)mdiArea->currentSubWindow();
				Datafont.setPointSize(Datafntsz.toInt());
				wr->setFont(Datafont);
			}
			delete dbuf;
		}
	}
	else
		retwaarde = 2;
	for(k = 0;k < p;k++)
		delete qn[k];
	delete d64Dialog;
	delete d64;
	d64Dialog = NULL;
	return retwaarde;
}

void MainWindowImpl::fileRename(QTableWidgetItem* renamed)
{
	if(renamed->column() == 0)
	{
		QString newName = renamed->text();
		char* oldname = DirNamePtr[renamed->row()];
		int i = newName.length();
		if(i != 0)
		{
			strncpy(oldname,newName.toLocal8Bit().constData(),i);
			for(;i < 16;i++)
				oldname[i] = 160;
			FileNameChanged = true;
		}
	}
}

bool MainWindowImpl::SubLoad(QString fn,QString type)
{
	QFont fonta;
	int fntsz;
	MdiChild* child = createMdiChild();
	if(type == BasicType)
	{
		if(TxtLineHighlight == "1")
			child->HighlightTxt = true;
		else
			child->HighlightTxt = false;
		child->InitLineNumber();
		child->setTabStopWidth(Tabspace.toInt() * (Txtfntsz.toInt() - 1));
		if(!child->loadFile(fn))
			return false;
		fntsz = Txtfntsz.toInt();
		fonta = Txtfont;
	}
	else if(type == PRGType)
	{
		curFile = QFileInfo(fn).canonicalFilePath();
		if(!child->loadPRG(fn))
			return false;
		fntsz = Prgfntsz.toInt();
		fonta = PRGfont;
	}
	else
	{
		if(!child->loadData(fn))
			return false;
		fntsz = Datafntsz.toInt();
		fonta = Datafont;
	}
	SetupChild(child);
	QMdiSubWindow* wr = (QMdiSubWindow*)mdiArea->currentSubWindow();
	fonta.setPointSize(fntsz);
	wr->setFont(fonta);
	return true;
}

void MainWindowImpl::SetupChild(MdiChild* child)
{
	if(mdiArea->subWindowList().count() == 1)
		child->showMaximized();
	else
	{
		child->showNormal();
		if(mdiArea->subWindowList().count() == 2)
		{
			QSize sz = mdiArea->size();
			mdiArea->cascadeSubWindows();
			foreach(QMdiSubWindow *window, mdiArea->subWindowList())
				window->resize((sz.width()*3)/4,(sz.height()*3)/4);
		}
	}
	connect(child,SIGNAL(cursorPositionChanged()),this,
		SLOT(DeCursorPositionChanged()));
}

void MainWindowImpl::DeCursorPositionChanged()
{
	QMdiSubWindow *window = mdiArea->activeSubWindow();
	if(window != NULL)
	{
		MdiChild *mdiChild = qobject_cast<MdiChild *>(window->widget());
		QTextCursor cursor = mdiChild->textCursor();
		QString pos;
		pos.setNum(cursor.positionInBlock());
		ColumnCursor->setText(pos);
	}
}

int MainWindowImpl::Renumber()
{
	bool ok = true;
	QMdiSubWindow *window = mdiArea->activeSubWindow();
	if(window)
	{
		QString tbuf;
		ushort erline;
		MdiChild *mdiChild = qobject_cast<MdiChild *>(window->widget());
		RenDlg = new QDialog();
		Ui_Renumber rn;
		rn.setupUi(RenDlg);
		int plines = mdiChild->document()->lineCount() - 1;
		if(mdiChild->document()->findBlockByNumber(plines).text().contains(QChar(' ')))
			plines++;
		if(mdiChild->lines != NULL)
			erline = mdiChild->lines[0].OrignalLineNumber;
		else
			erline = 1;
		rn.spinBox_2->setValue(erline);
		rn.spinBox_3->setValue(erline);
		tbuf = mdiChild->document()->findBlockByNumber(plines-1).text();
		int LastLine = tbuf.left(tbuf.indexOf(" ")).toInt();
		rn.spinBox_4->setValue(LastLine);
		RenDlg->setWindowFlags(Qt::WindowStaysOnTopHint);
		RenDlg->show();
		RenDlg->exec();
		if(RenDlg->result())
		{
			QString uitbuf("");
			ushort inc = (ushort)rn.spinBox->value();
			ushort bgnline = (ushort)rn.spinBox_2->value();
			ushort bline = (ushort)rn.spinBox_3->value();
			ushort endline = (ushort)rn.spinBox_4->value() + 1;
			if(endline == 0)
				endline = 65535;
			ushort nxtline = bgnline - inc;
			int otest;
			ushort* linerel = new ushort [plines];
			ushort* linetran = new ushort [plines];
			int i = 0;
			for(;i < plines;i++)
			{
				tbuf = mdiChild->document()->findBlockByNumber(i).text();
				linerel[i] = tbuf.left(tbuf.indexOf(" ")+1).toUShort(&ok);
				if(!ok)
				{
					if(i > 0)
						erline = linerel[i-1];
					else
						erline = 1;
					delete linetran;
					delete linerel;
					return erline;
				}
				if((linerel[i] <= bline) || (linerel[i] >= endline))
				{
					if((bgnline > bline) && (endline == LastLine + 1))
						linetran[i] = nxtline + inc;
					else
						linetran[i] = linerel[i];
					nxtline = linetran[i];
				}
				else
				{
					otest = nxtline + inc;
					if(((otest > endline) && (endline < LastLine)) || (otest > 65535))
					{
						if(i > 0)
							erline = linerel[i-1];
						else
							erline = 1;
						delete linetran;
						delete linerel;
						return erline;
					}
					nxtline += inc;
					linetran[i] = nxtline;
				}
			}
			PRGLines* mlines = new PRGLines[i+10];
			mdiChild->ResizePRGLines(i+10);
			ushort pl;
			ushort ofst = 2049;
			for(i = 0;i < plines;i++)
			{
				mlines[i].OrignalLineNumber = linetran[i];
				tbuf = mdiChild->document()->findBlockByNumber(i).text();
				mlines[i].TranslatedLine = tbuf.right(tbuf.length()-tbuf.indexOf(" ")-1);
				pl= MakeLocal(i,linetran,linerel,mlines,plines);
				if(pl == 65535)
				{
					delete[] mlines;
					delete linetran;
					delete linerel;
					return lastChecked;
				}
				ofst += pl;
				mlines[i].OrignalLineOffset = ofst;
			}
			for(i = 0;i < plines;i++)
			{
				tbuf.sprintf("%d ",mlines[i].OrignalLineNumber);
				uitbuf += tbuf;
				strcpy((char*)mdiChild->lines[i].OrignalLine,
					(char*)mlines[i].OrignalLine);
				mdiChild->lines[i].TranslatedLine = mlines[i].TranslatedLine;
				mdiChild->lines[i].OrignalLineNumber = mlines[i].OrignalLineNumber;
				mdiChild->lines[i].OrignalLineOffset =
					mdiChild->lines[i].OrignalLineOffset;
				uitbuf += mlines[i].TranslatedLine;
				uitbuf += "\n";
			}
			mdiChild->setPlainText(uitbuf);
			mdiChild->setWordWrapMode(QTextOption::NoWrap);
			mdiChild->document()->setModified(true);
			delete[] mlines;
			delete linetran;
			delete linerel;
		}
		delete RenDlg;
		RenDlg = NULL;
	}
	return 0;
}

MdiChild* MainWindowImpl::createMdiChild()
{
	MdiChild *child = new MdiChild;
	mdiArea->addSubWindow(child);
	return child;
}

int MainWindowImpl::RenumberEx(int jx,int ix,ushort* xfrstln,int mx,
	ushort* xline,const uchar* dx,PRGLines* mxlines,int pxlin)
{
	int p;
	int n = jx;
	uchar gnum[8];
	for(p = 0;(dx[n] > 47) && (dx[n] < 58);n++,p++)
		gnum[p] = dx[n];
	gnum[p] = 0;
	int r;
	lastChecked = (ushort)atoi((char*)gnum);
	for(r = 0;r < pxlin;r++)
	{
		if(lastChecked == xline[r])
		{
			sprintf((char*)mxlines[ix].OrignalLine + mx,"%d",xfrstln[r]);
			mxlines[ix].TranslatedLine.replace(jx,p,
				(char*)(mxlines[ix].OrignalLine + mx));
			for(n = 0;mxlines[ix].OrignalLine[mx] != 0;mx++,n++);
			break;
		}
	}
	if(r == pxlin)
		return -1;
	return n;
}

QMdiSubWindow* MainWindowImpl::findMdiChild(const QString &fileName)
{
	QString canonicalFilePath = QFileInfo(fileName).canonicalFilePath();
	foreach(QMdiSubWindow *window, mdiArea->subWindowList())
	{
		MdiChild *mdiChild = qobject_cast<MdiChild *>(window->widget());
		if(mdiChild->currentFile() == canonicalFilePath)
			return window;
	}
	return 0;
}

bool MainWindowImpl::event(QEvent* ev)
{
	if(ev->type() == QEvent::Resize)
	{
		QSize nsize = size(); 
		if(bsize != nsize)
		{
			bsize = nsize;
			nsize.rheight() -= 115;
			nsize.rwidth() -= 22;
			mdiArea->resize(nsize);
		}
	}
	else if((ev->type() == KeyPres))
	{
		QKeyEvent* kev = (QKeyEvent*)ev;
		switch(kev->key())
		{
			case Qt::Key_F3:
				if(fndtxt != "")
				{
					QMdiSubWindow *window = mdiArea->activeSubWindow();
					if(window != NULL)
					{
						MdiChild *mdiChild = qobject_cast<MdiChild *>(window->widget());
						mdiChild->find(fndtxt,(QTextDocument::FindFlags)prst);
					}
				}
				break;
			case Qt::Key_CapsLock:	//this does't work in windows always
				Capslok = !Capslok;	//do something else for windows.
				if(Capslok)
					Capslock->setText("<font color='red'>CAPS</font>");
				else
					Capslock->setText("<font color='black'>CAPS</font>");
				break;
			case Qt::Key_NumLock:
				Numlok = !Numlok;
				if(Numlok)
					Numlock->setText("<font color='red'>NUM</font>");
				else
					Numlock->setText("<font color='black'>NUM</font>");
				break;
			case Qt::Key_ScrollLock:
				Scrolok = !Scrolok;
				if(Scrolok)
					Scrolllock->setText("<font color='red'>SCROL</font>");
				else
					Scrolllock->setText("<font color='black'>SCROL</font>");
				break;
		}
	}
	else if(ev->type() == QEvent::Close)
	{
		if(FndDlg != NULL)			//close all the dialog boxes that are open
			FndDlg->accept();			//when ending the program
		if(AboutDialog != NULL)
			AboutDialog->accept();
		if(Aboutqt != NULL)
			Aboutqt->accept();
		if(CfgDialog != NULL)
			CfgDialog->accept();
		if(InsDialog != NULL)
			InsDialog->accept();
		if(fnlog != NULL)
			fnlog->accept();
		if(GotoDlg != NULL)
			GotoDlg->accept();
		if(d64Dialog != NULL)
			d64Dialog->accept();
		if(RenDlg != NULL)
			RenDlg->accept();
		if(AsmResDialog != NULL)
			AsmResDialog->accept();
	}
	return QMainWindow::event(ev);
}

#ifndef WIN_32
bool MainWindowImpl::keyState(int iKey,Display *pDisplay)
{
	int		iKeyMask = 0;
	Window	wDummy1, wDummy2;
	int		iDummy3, iDummy4, iDummy5, iDummy6;
 	unsigned int iMask;
 	XModifierKeymap* map = XGetModifierMapping(pDisplay);
 	KeyCode keyCode = XKeysymToKeycode(pDisplay,iKey);
	for(int i = 0; i < 8; ++i)
	{
		if( map->modifiermap[map->max_keypermod * i] == keyCode)
			iKeyMask = 1 << i;
	}
	XQueryPointer(pDisplay,DefaultRootWindow(pDisplay),&wDummy1,&wDummy2,
		&iDummy3,&iDummy4,&iDummy5,&iDummy6,&iMask );
 	XFreeModifiermap(map);
	return (iMask & iKeyMask) != 0;
}
#endif

ushort MainWindowImpl::MakeLocal(int i,ushort* linetran,ushort* linerel,
	PRGLines* mlines,int pxlin)
{
	int t;
	int m = 0;
	uchar dline[96];
	for(t = 0;t < mlines[i].TranslatedLine.length();t++)
		dline[t] = (uchar)(mlines[i].TranslatedLine.at(t).unicode() & 255);
	dline[t] = 0;
	for(int j = 0;dline[j] != 0;)
	{
		switch(dline[j])
		{
			case '"':
				do
				{
					mlines[i].OrignalLine[m++] = dline[j++];
					if(dline[j] == 0)
					{
						mlines[i].OrignalLine[m] = 0;  
						return ushort(m+5);
					}
				}while(dline[j] != '"');
				mlines[i].OrignalLine[m++] = dline[j++];
				break;
			case 'A':
				switch(dline[j+1])
				{
					case 'B':
						if(dline[j+2] == 'S')
						{
							mlines[i].OrignalLine[m++] = 182;
							j += 3;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					case 'N':
						if(dline[j+2] == 'D')
						{
							mlines[i].OrignalLine[m++] = 175;
							j += 3;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					case 'S':
						if(dline[j+2] == 'C')
						{
							mlines[i].OrignalLine[m++] = 198;
							j += 3;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					case 'T':
						if(dline[j+2] == 'N')
						{
							mlines[i].OrignalLine[m++] = 193;
							j += 3;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					default:
						mlines[i].OrignalLine[m++] = dline[j++];
						break;
				}
			break;
			case 'C':
				switch(dline[j+1])
				{
					case 'H':
						if((dline[j+2] == 'R') && (dline[j+3] == '$'))
						{
							mlines[i].OrignalLine[m++] = 199;
							j += 4;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					case 'M':
						if(dline[j+2] == 'D')
						{
							mlines[i].OrignalLine[m++] = 157;
							j += 3;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					case 'L':
						if(dline[j+2] == 'R')
						{
							mlines[i].OrignalLine[m++] = 156;
							j += 3;
						}
						else if((dline[j+2] == 'O') && (dline[j+3] == 'S') &&
							(dline[j+4] == 'E'))
						{
							mlines[i].OrignalLine[m++] = 160;
							j += 5;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					case 'O':
						if(dline[j+2] == 'S')
						{
							mlines[i].OrignalLine[m++] = 190;
							j += 3;
						}
						else if((dline[j+2] == 'N') && (dline[j+3] == 'T'))
						{
							mlines[i].OrignalLine[m++] = 154;
							j += 4;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					default:
						mlines[i].OrignalLine[m++] = dline[j++];
						break;
				}
				break;
			case 'D':
				switch(dline[j+1])
				{
					case 'A':
						if((dline[j+2] == 'T') && (dline[j+3] == 'A'))
						{
							mlines[i].OrignalLine[m++] = 131;
							j += 4;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					case 'E':
						if(dline[j+2] == 'F')
						{
							mlines[i].OrignalLine[m++] = 150;
							j += 3;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					case 'I':
						if(dline[j+2] == 'M')
						{
							mlines[i].OrignalLine[m++] = 134;
							j += 3;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					default:
						mlines[i].OrignalLine[m++] = dline[j++];
						break;
				}
				break;
			case 'E':
				if((dline[j+1] == 'N') && (dline[j+2] == 'D'))
				{
					mlines[i].OrignalLine[m++] = 128;
					j += 3;
				}
				else
					mlines[i].OrignalLine[m++] = dline[j++];
				break;
				case 'F':
				switch(dline[j+1])
				{
					case 'N':
						mlines[i].OrignalLine[m++] = 165;
						j+=2;
						break;
					case 'O':
						if(dline[j+2] == 'R')
						{
							mlines[i].OrignalLine[m++] = 129;
							j += 3;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					case 'R':
						if(dline[j+2] == 'E')
						{
							mlines[i].OrignalLine[m++] = 184;
							j += 3;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					default:
						mlines[i].OrignalLine[m++] = dline[j++];
						break;
				}
				break;
			case 'G':
				if((dline[j+1] == 'E') && (dline[j+2] == 'T'))
				{
					mlines[i].OrignalLine[m++] = 161;
					j += 3;
				}
				else if(dline[j+1] == 'O')
				{
					if((dline[j+2] == 'T') && (dline[j+3] == 'O'))
					{
						mlines[i].OrignalLine[m++] = 137;
						j+=4;
						while(dline[j] == ' ')
							mlines[i].OrignalLine[m++] = dline[j++];
						if(linetran != NULL)
						{
							t = RenumberEx(j,i,linetran,m,linerel,dline,
								mlines,pxlin);
							if(t == -1)
								break;
							j += t;
							m += t;
							for(t = 0;t < mlines[i].TranslatedLine.length();t++)
								dline[t]=(uchar)(mlines[i].TranslatedLine.at(t).unicode()&255);
							dline[t] = 0;
							while(dline[j] == ',')
							{
								mlines[i].OrignalLine[m++] = dline[j++];
								t = RenumberEx(j,i,linetran,m,linerel,dline,
									mlines,pxlin);
								if(t == -1)
									break;
								j += t;
								m += t;
								for(t = 0;t < mlines[i].TranslatedLine.length();t++)
									dline[t]=(uchar)(mlines[i].TranslatedLine.at(t).unicode()&255);
								dline[t] = 0;
							}
						}
						else
						{
							do
							{
								if(dline[j] == 0)
									break;
								mlines[i].OrignalLine[m++] = dline[j++];
							}while(((dline[j-1] > 47) && (dline[j-1] < 58)) ||
								(dline[j-1] == ',') || (dline[j-1] == ' '));
						}
					}
					else if((dline[j+2] == 'S') && (dline[j+3] == 'U') &&
						(dline[j+4] == 'B'))
					{
						mlines[i].OrignalLine[m++] = 141;
						j+=5;
						while(dline[j] == ' ')
							mlines[i].OrignalLine[m++] = dline[j++];
						if(linetran != NULL)
						{
							t = RenumberEx(j,i,linetran,m,linerel,dline,
								mlines,pxlin);
							if(t == -1)
								break;
							j += t;
							m += t;
							for(t = 0;t < mlines[i].TranslatedLine.length();t++)
								dline[t] =
									(uchar)(mlines[i].TranslatedLine.at(t).unicode()&255);
							dline[t] = 0;
							while(dline[j] == ',')
							{
								mlines[i].OrignalLine[m++] = dline[j++];
								t = RenumberEx(j,i,linetran,m,linerel,dline,
									mlines,pxlin);
								if(t == -1)
									break;
								j += t;
								m += t;
								for(t = 0;t < mlines[i].TranslatedLine.length();t++)
									dline[t] =
										(uchar)(mlines[i].TranslatedLine.at(t).unicode()&255);
								dline[t] = 0;
							}
						}
						else
						{
							do
							{
								if(dline[j] == 0)
									break;
								mlines[i].OrignalLine[m++] = dline[j++];
							}while(((dline[j-1] > 47) && (dline[j-1] < 58)) ||
								(dline[j-1] == ',') || (dline[j-1] == ' '));
						}
					}
					else
						mlines[i].OrignalLine[m++] = dline[j++];
				}
				else
					mlines[i].OrignalLine[m++] = dline[j++];
				break;
			case 'I':
				if(dline[j+1] == 'F')
				{
					mlines[i].OrignalLine[m++] = 139;
					j += 2;
				}
				else if(dline[j+1] == 'N')
				{
					if(dline[j+2] == 'T')
					{
						mlines[i].OrignalLine[m++] = 181;
						j += 3;
					}
					else if((dline[j+2] == 'P') && (dline[j+3] == 'U') &&
						(dline[j+4] == 'T'))
					{
						if(dline[j+5] == '#')
						{
							mlines[i].OrignalLine[m++] = 132;
							j += 6;
						}
						else
						{
							mlines[i].OrignalLine[m++] = 133;
							j += 5;
						}
					}
					else
						mlines[i].OrignalLine[m++] = dline[j++];
				}
				else
					mlines[i].OrignalLine[m++] = dline[j++];
				break;
			case 'L':
				switch(dline[j+1])
				{
					case 'E':
						switch(dline[j+2])
						{
							case 'T':
								mlines[i].OrignalLine[m++] = 136;
								j += 3;
								break;
							case 'N':
								mlines[i].OrignalLine[m++] = 195;
								j += 3;
								break;
							case 'F':
								if((dline[j+3] == 'T') && (dline[j+4] == '$'))
								{
									mlines[i].OrignalLine[m++] = 200;
									j += 5;
								}
								else
									mlines[i].OrignalLine[m++] = dline[j++];
								break;
							default:
								mlines[i].OrignalLine[m++] = dline[j++];
								break;
						}
						break;
					case 'I':
						if((dline[j+2] == 'S') && (dline[j+3] == 'T'))
						{
							mlines[i].OrignalLine[m++] = 155;
							j += 4;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					case 'O':
						if((dline[j+2] == 'A') && (dline[j+3] == 'D'))
						{
							mlines[i].OrignalLine[m++] = 147;
							j += 4;
						}
						else if(dline[j+2] == 'G')
						{
							mlines[i].OrignalLine[m++] = 188;
							j += 3;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					default:
						mlines[i].OrignalLine[m++] = dline[j++];
						break;
				}
			break;
			case 'M':
				if((dline[j+1] == 'I') && (dline[j+2] == 'D')
					&& (dline[j+3] == '$'))
				{
					mlines[i].OrignalLine[m++] = 202;
					j += 4;
				}
				else
					mlines[i].OrignalLine[m++] = dline[j++];
				break;
			case 'N':
				if((dline[j+1] == 'O') && (dline[j+2] == 'T'))
				{
					mlines[i].OrignalLine[m++] = 168;
					j += 3;
				}
				else if(dline[j+1] == 'E')
				{
					if(dline[j+2] == 'W')
					{
						mlines[i].OrignalLine[m++] = 162;
						j += 3;
					}
					else if((dline[j+2] == 'X')  && (dline[j+3] == 'T'))
					{
						mlines[i].OrignalLine[m++] = 130;
						j += 4;
					}
					else
						mlines[i].OrignalLine[m++] = dline[j++];
				}
				else
					mlines[i].OrignalLine[m++] = dline[j++];
				break;
			case 'O':
				switch(dline[j+1])
				{
					case 'N':
						mlines[i].OrignalLine[m++] = 145;
						j += 2;
						break;
					case 'P':
						if((dline[j+2] == 'E') && (dline[j+3] == 'N'))
						{
							mlines[i].OrignalLine[m++] = 159;
							j += 4;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					case 'R':
						mlines[i].OrignalLine[m++] = 176;
						j += 2;
						break;
					default:
						mlines[i].OrignalLine[m++] = dline[j++];
						break;
				}
			break;
			case 'P':
				switch(dline[j+1])
				{
					case 'E':
						if((dline[j+2] == 'E') && (dline[j+3] == 'K'))
						{
							mlines[i].OrignalLine[m++] = 194;
							j += 4;
						}
						else
						mlines[i].OrignalLine[m++] = dline[j++];
						break;
					case 'O':
						if((dline[j+2] == 'K') && (dline[j+3] == 'E'))
						{
							mlines[i].OrignalLine[m++] = 151;
							j += 4;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					case 'R':
						if((dline[j+2] == 'I') && (dline[j+3] == 'N')
							&& (dline[j+4] == 'T'))
						{
							if(dline[j+5] == '#')
							{
								mlines[i].OrignalLine[m++] = 152;
								j += 6;
							}
							else
							{
								mlines[i].OrignalLine[m++] = 153;
								j += 5;
							}
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					default:
						mlines[i].OrignalLine[m++] = dline[j++];
						break;
				}
			break;
			case 'R':
				switch(dline[j+1])
				{
					case 'E':
						switch(dline[j+2])
						{
							case 'A':
								if(dline[j+3] == 'D')
								{
									mlines[i].OrignalLine[m++] = 135;
									j += 4;
								}
								else
									mlines[i].OrignalLine[m++] = dline[j++];
								break;
							case 'M':
								mlines[i].OrignalLine[m++] = 143;
								j += 3;
								while(dline[j] != 0)
									mlines[i].OrignalLine[m++] = dline[j++];
								break;
							case 'S':
								if((dline[j+3] == 'T') && (dline[j+4] == 'O')
									&& (dline[j+5] == 'R') && (dline[j+6] == 'E'))
								{
									mlines[i].OrignalLine[m++] = 140;
									j += 7;
								}
								else
									mlines[i].OrignalLine[m++] = dline[j++];
								break;
							case 'T':
								if((dline[j+3] == 'U') && (dline[j+4] == 'R')
									&& (dline[j+5] == 'N'))
								{
									mlines[i].OrignalLine[m++] = 142;
									j += 6;
								}
								else
									mlines[i].OrignalLine[m++] = dline[j++];
								break;
							default:
								mlines[i].OrignalLine[m++] = dline[j++];
								break;
						}
						break;
					case 'I':
						if((dline[j+2] == 'G') && (dline[j+3] == 'H')
							&& (dline[j+4] == 'T') && (dline[j+5] == '$'))
						{
							mlines[i].OrignalLine[m++] = 201;
							j += 6;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					case 'N':
						if(dline[j+2] == 'D')
						{
							mlines[i].OrignalLine[m++] = 187;
							j += 3;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					case 'U':
						if(dline[j+2] == 'N')
						{
							mlines[i].OrignalLine[m++] = 138;
							j += 3;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					default:
						mlines[i].OrignalLine[m++] = dline[j++];
						break;
				}
				break;
			case 'S':
				switch(dline[j+1])
				{
					case 'A':
						if((dline[j+2] == 'V') && (dline[j+3] == 'E'))
						{
							mlines[i].OrignalLine[m++] = 148;
							j += 4;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					case 'G':
						if(dline[j+2] == 'N')
						{
							mlines[i].OrignalLine[m++] = 180;
							j += 3;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					case 'I':
						if(dline[j+2] == 'N')
						{
							mlines[i].OrignalLine[m++] = 191;
							j += 3;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					case 'P':
						if((dline[j+2] == 'C') && (dline[j+3] == '('))
						{
							mlines[i].OrignalLine[m++] = 166;
							j +=4;											
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					case 'Q':
						if(dline[j+2] == 'R')
						{
							mlines[i].OrignalLine[m++] = 186;
							j += 3;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					case 'T':
						if((dline[j+2] == 'E') && (dline[j+3] == 'P'))
						{
							mlines[i].OrignalLine[m++] = 169;
							j +=4;
						}
						else if((dline[j+2] == 'O') && (dline[j+3] == 'P'))
						{
							mlines[i].OrignalLine[m++] = 144;
							j +=4;
						}
						else if((dline[j+2] == 'R') && (dline[j+3] == '$'))
						{
							mlines[i].OrignalLine[m++] = 196;
							j +=4;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					case 'Y':
						if(dline[j+2] == 'S')
						{
							mlines[i].OrignalLine[m++] = 158;
							j += 3;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					default:
						mlines[i].OrignalLine[m++] = dline[j++];
						break;
				}
				break;
			case 'T':
				switch(dline[j+1])
				{
					case 'A':
						if((dline[j+2] == 'B') && (dline[j+3] == '('))
						{
							mlines[i].OrignalLine[m++] = 163;
							j+=4;
						}
						else if(dline[j+2] == 'N')
						{
							mlines[i].OrignalLine[m++] = 192;
							j+=3;
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					case 'H':
						if((dline[j+2] == 'E') && (dline[j+3] == 'N'))
						{
							mlines[i].OrignalLine[m++] = 167;
							j+=4;
							while(dline[j] == ' ')
								mlines[i].OrignalLine[m++] = dline[j++];
							if(linetran != NULL)
							{
								if((dline[j] > 47) && (dline[j] < 58))
								{
									t = RenumberEx(j,i,linetran,m,
										linerel,dline,mlines,pxlin);
									if(t == -1)
										break;
									j += t;
									m += t;
									for(t = 0;t < mlines[i].TranslatedLine.length();t++)
										dline[t] =
											(uchar)(mlines[i].TranslatedLine.at(t).unicode()&255);
									dline[t] = 0;
								}
							}
							else
							{
								while(((dline[j] > 47) && (dline[j] < 58)) ||
									(dline[j] == ',') || (dline[j] == ' '))
									mlines[i].OrignalLine[m++] = dline[j++];
							}
						}
						else
							mlines[i].OrignalLine[m++] = dline[j++];
						break;
					case 'O':
						mlines[i].OrignalLine[m++] = 164;
						j+=2;
						break;
					default:
						mlines[i].OrignalLine[m++] = dline[j++];
						break;
				}
				break;
			case 'U':
				if((dline[j+1] == 'S') && (dline[j+2] == 'R'))
				{
					mlines[i].OrignalLine[m++] = 183;            
					j+=3;
				}
				else
					mlines[i].OrignalLine[m++] = dline[j++];
				break;
			case 'V':
				if((dline[j+1] == 'A') && (dline[j+2] == 'L'))
				{
					mlines[i].OrignalLine[m++] = 197;
					j+=3;
				}
				else if((dline[j+1] == 'E') && (dline[j+2] == 'R') &&
					(dline[j+3] == 'I') && (dline[j+4] == 'F') && (dline[j+5] == 'Y'))
				{
					mlines[i].OrignalLine[m++] = 149;
					j += 6;
				}
				else
					mlines[i].OrignalLine[m++] = dline[j++];
				break;
			case 'W':
				if((dline[j+1] == 'A') && (dline[j+2] == 'I') &&
					(dline[j+3] == 'T'))
				{
					mlines[i].OrignalLine[m++] = 146;
					j += 4;
				}
				else
					mlines[i].OrignalLine[m++] = dline[j++];
				break;
			case '+':
				mlines[i].OrignalLine[m++] = 170;
				j++;
				break;
			case '-':
				mlines[i].OrignalLine[m++] = 171;
				j++;
				break;
			case '*':
				mlines[i].OrignalLine[m++] = 172;
				j++;
				break;
			case '/':
				mlines[i].OrignalLine[m++] = 173;
				j++;
				break;
			case '^':
				mlines[i].OrignalLine[m++] = 174;
				j++;
				break;
			case '>':
				mlines[i].OrignalLine[m++] = 177;
				j++;
				break;
			case '=':
				mlines[i].OrignalLine[m++] = 178;
				j++;
				break;
			case '<':
				mlines[i].OrignalLine[m++] = 179;
				j++;
				break;
			default:
				mlines[i].OrignalLine[m++] = dline[j++];
				break;
		}
	}
	if(t == -1)
		return 65535;
	mlines[i].OrignalLine[m] = 0;  
	return ushort(m+5);
}
